import argparse
import glob
import os

import matplotlib
import pandas as pd

matplotlib.use('Agg')
from matplotlib import pyplot as plt


class TimeSeriesAggregator:
    """
    Main object of the class. Stores many relevant values and data frames used for plotting
    """
    def __init__(self):
        self.parser = None
        self.input_path = None
        self.output_path = None
        self.marker_pos_x = None
        self.label_text = None
        self.agg_latency = None
        self.agg_throughput = None
        self.cols_latency = None
        self.cols_throughput = None
        self.run = -1
        self.files = 0
        self.min_t = 0
        self.max_t = 0
        self.min_l = 0
        self.max_l = 0
        self.plotAverage = False

    def parser_setup(self):
        """
        sets up a parser that allows to read parameters when executing
        """

        #TODO: add output format as parameter: pdf/png and align plotting call

        # define input directory with --input/-i and the output path with --output/-o on which
        # the resulting plots shall be stored
        parser = argparse.ArgumentParser(description='Plot timeseries based on YCSB result files.')
        parser.add_argument('--input', '-i', type=str, nargs=1, required=True,
                            help='path of the folder with the load.txt')
        #parser.add_argument('--output', '-o', type=str, nargs=1, required=True,
        #                    help='path to the store timeseries plot')
        # optional arguments for marker position at the x axis and the label_text if a marker is given
        parser.add_argument('--marker_pos_x', '-pos', type=int, nargs=1, default=None, required=False,
                            help='fixing a marker at given x position')
        parser.add_argument('--label', '-l', type=str, nargs=1, default=None, required=False,
                            help='label text to display at the marker position (argument marker_pos_x')
        # optional argument: run number (the specified run will be shown in the whole context)
        parser.add_argument('--run', '-r', type=int, choices=range(0, 10), default=-1, required=False)
        parser.add_argument('--plotAverage', '-pa', type=bool, default=False, required=False)
        # fix the parser as main parser object of the TimeSeriesAggregator object
        self.parser = parser

    def parse_input(self):
        """
        sets the essential values input_path, output_path, marker_pos_x and label_text for later plotting
        """
        # process input parameters
        args = self.parser.parse_args()
        # define inputPath as the directory with load.txt files
        self.input_path = args.input[0]
        
        self.output_path = self.input_path
        # create, if necessary a new directory with the given output path (creates as many subdirectories as necessary)
        #self.output_path = args.output[0]     
        #if not os.path.isdir(self.output_path):
        #    os.makedirs(self.output_path)
        # set the marker and label_text if given, otherwise None
        if args.marker_pos_x is not None:
            self.marker_pos_x = args.marker_pos_x[0]
        else:
            self.marker_pos_x = None
        if args.label is not None:
            self.label_text = args.label[0]
        else:
            self.label_text = None
        # get the run number, if given
        if args.run is not None:
            self.run = args.run

        if args.plotAverage:
            self.plotAverage = True


    def extract_input(self):
        """
        data is prepared and processed for later plotting
        """
        # get all files from the given directory
        all_files = glob.glob(os.path.join(self.input_path, "*.txt"))
        file_list = []
        # attribute that determines how many timestamp-rows will be created
        current_max = 0
        # create one data frame for each read file
        for filename in all_files:
            df = pd.read_csv(filename, error_bad_lines=False, sep=";",
                             names=['timeseries', 'timestamp', 'throughput', 'latency'],
                             index_col=False)
            file_list.append(df)
            self.files = len(file_list)

            # determine the current maximum of timestamps (fixes the length of the data frame later)
            maximum_timestamp_of_file = max(df['timestamp'])
            if maximum_timestamp_of_file > current_max:
                current_max = maximum_timestamp_of_file

        # create an empty data frame with only timestamps for latency and throughput
        self.cols_latency = ['timestamp'] + ['l{}'.format(i) for i in range(len(all_files))]
        self.cols_throughput = ['timestamp'] + ['t{}'.format(i) for i in range(len(all_files))]
        agg_frame_latency = pd.DataFrame(columns=self.cols_latency)
        agg_frame_latency['timestamp'] = range(10, int(current_max) + 10, 10)
        agg_frame_throughput = pd.DataFrame(columns=self.cols_throughput)
        agg_frame_throughput['timestamp'] = range(10, int(current_max) + 10, 10)

        # fill the new data frames, enumerate columns with file numbers
        for index, file in enumerate(file_list):
            agg_frame_latency['l{}'.format(index)] = file['latency']
            agg_frame_throughput['t{}'.format(index)] = file['throughput']

        # change the index so that timestamps are the index
        agg_frame_latency.set_index('timestamp', inplace=True)
        agg_frame_throughput.set_index('timestamp', inplace=True)
        # calculate mean for each row
        agg_frame_latency['Mean'] = agg_frame_latency.mean(axis=1, skipna=True)
        agg_frame_throughput['Mean'] = agg_frame_throughput.mean(axis=1, skipna=True)

        indices_latency = [x for x in agg_frame_latency.columns if x != "Mean"]
        indices_throughput = [x for x in agg_frame_throughput.columns if x != "Mean"]

        # the factor of the standard deviation is 1/n-1, thus it is clearly specified as ddof=1
        # calculate standard deviation per row
        agg_frame_latency['Standard Deviation'] = agg_frame_latency[indices_latency].std(axis=1, skipna=True, ddof=1)
        agg_frame_throughput['Standard Deviation'] = agg_frame_throughput[indices_throughput].std(axis=1, skipna=True, ddof=1)
        # save results in self-object
        self.agg_latency = agg_frame_latency
        self.agg_throughput = agg_frame_throughput

    def plot_all_timeseries(self):
        """
        processes the collected data and generates plots out of the given data
        """

        # THROUGHPUT
        # indicate the start of the plotting process
        print("Plotting timeseries data...")


        #TODO: only clean outliers if values are really outliers, cleaning all values by default will corrupt the data, i.e. determination of an outliers needs to be defined 
        #TODO: move to its own function 
        # cleaning of outliers
        #for column in self.agg_throughput.columns:
        #    self.agg_throughput = self.remove_outliers(self.agg_throughput, column)
        # determine the min and max values of throughput to adjust the scaling of the graphs later


        self.min_t = self.agg_throughput.iloc[:, :-2].min().min()
        self.max_t = self.agg_throughput.iloc[:, :-2].max().max()

        timestamps_t = self.agg_throughput.index
        ax = self.agg_throughput.plot()
        # limit the area shown in the plot, depending on min and max values, adjusted by 300 for better view on the plot
        plt.ylim(self.min_t - 300, self.max_t + 300)
        #plt.ylim(0, self.max_t + 300)
        # remove the line for standard deviation as it is painted as an area around the mean
        ax.lines.remove(ax.lines[-1])
        if self.run != -1:
            for i in range(self.files, -1, -1):
                if i != self.run:
                    ax.lines.remove(ax.lines[i])
        # calculate the overall mean
        overall_mean = self.agg_throughput["Mean"].mean()
        # draw the overall mean as horizontal line if average flag is true
        if self.plotAverage:
            ax.axhline(overall_mean, linestyle='--', label='Overall Average', color='black')
        # draw the marker with its label on the graph if given
        if self.marker_pos_x is not None:
            ax.axvline(self.marker_pos_x, color='red', linestyle=':', label=self.label_text)
        # save file under the predetermined directory
        output_file = os.path.join(self.output_path + "throughput-ts.pdf")
        # draw the standard deviation as area around the mean
        mean_t = self.agg_throughput["Mean"]
        std_t = self.agg_throughput["Standard Deviation"]
        plt.fill_between(timestamps_t, mean_t - 2 * std_t,
                         mean_t + 2 * std_t, color='b', alpha=0.2, label='Standard Deviation')
        # modify default legend
        ax.legend(loc='lower center', ncol=3, borderpad=1)
        # adjust size of the figure
        plt.plot(figsize=(1000, 600))
        # indicating saving process
        print("Saving throughput plot to disk: " + output_file)
        # store the created picture
        plt.savefig(output_file, format='pdf')
        plt.close()

        # LATENCY
        #TODO: move latency plot in its own function

        #TODO: only clean outliers if values are really outliers, cleaning all values by default will corrupt the data, i.e. determination of an outliers needs to be defined 
        #TODO: move to its own function 
        # cleaning of outliers
        # cleaning of outliers
        #for column in self.agg_latency.columns:
        #    self.agg_latency = self.remove_outliers(self.agg_latency, column)


        # determine the min and max values of latency to adjust the scaling of the graphs later
        self.min_l = self.agg_latency.iloc[:, :-2].min().min()
        self.max_l = self.agg_latency.iloc[:, :-2].max().max()
        timestamps_l = self.agg_latency.index
        ax = self.agg_latency.plot()
        # limit the area shown in the plot, depending on min and max values, adjusted by 300 for better view on the plot
        plt.ylim(self.min_l - 100, self.max_l + 100)
        # remove the line for standard deviation as it is painted as an area around the mean
        ax.lines.remove(ax.lines[-1])
        if self.run != -1:
            for i in range(self.files, -1, -1):
                if i != self.run:
                    ax.lines.remove(ax.lines[i])
        # calculate the overall mean
        overall_mean = self.agg_latency["Mean"].mean()
        # draw the overall mean as horizontal line
        if self.plotAverage:
            ax.axhline(overall_mean, linestyle='--', label='Overall Average', color='black')
        # draw the marker with its label on the graph if given
        if self.marker_pos_x is not None:
            ax.axvline(self.marker_pos_x, color='red', linestyle=':', label=self.label_text)
        # save file under the predetermined directory
        output_file = os.path.join(self.output_path + "latency-ts.pdf")
        # draw the standard deviation as area around the mean
        mean_l = self.agg_latency["Mean"]
        std_l = self.agg_latency["Standard Deviation"]
        plt.fill_between(timestamps_l, mean_l - 2 * std_l,
                         mean_l + 2 * std_l, color='b', alpha=0.2, label='Standard Deviation')
        # modify default legend
        ax.legend(loc='lower center', ncol=3, borderpad=1)
        # adjust size of the figure
        plt.plot(figsize=(1000, 600))
        # indicating saving process
        print("Saving latency plot to disk: " + output_file)
        # store the created picture
        plt.savefig(output_file, format='pdf')

    @staticmethod
    def remove_outliers(df_in, col_name):
        """
        removes data point outliers from the set and adjusts them with help of the interquartile range
        :param df_in: the input data frame from which shall be cleaned
        :param col_name: the current column name: in which column the calculation shall be performed
        :return: a cleaned data frame
        """
        q1 = df_in[col_name].quantile(0.25)
        q3 = df_in[col_name].quantile(0.75)
        iqr = q3 - q1  # Interquartile range
        fence_low = q1 - 1.5 * iqr
        fence_high = q3 + 1.5 * iqr
        df_out = df_in.loc[(df_in[col_name] > fence_low) & (df_in[col_name] < fence_high)]
        return df_out

    def process(self):
        """
        main method. Executes the methods in correct order and terminates after running
        """
        self.parser_setup()
        self.parse_input()
        self.extract_input()
        self.plot_all_timeseries()


if __name__ == "__main__":
    TimeSeriesAggregator().process()
