#! python3
import os
import numpy as np
import json
import pprint
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import argparse
import pandas as pd
import plotting_config




class TimeseriesData:
    timestamps = []
    operations = []
    latencies = []
    scaleOutAverage = None



def plot_timeseries():

    parser = argparse.ArgumentParser(description='Plot timeseries and averages of different phases based on YCSB result files.')
    parser.add_argument('--runFolder', '-rf',type=str, nargs=1,required=True, help='path to the evalation_run_XXX folder')
    parser.add_argument('--scenario', '-s',type=str, nargs=1,required=True, choices=['elasticity', 'availability'], help='type of the evaluation scenario: elasticity/availability')
    parser.add_argument('--scaleDelay', '-sd',type=int, nargs='?',const=0, help='delay in seconds before scaling action was triggered')
    parser.add_argument('--gibbonDelay', '-gd',type=int, nargs='?',const=0, help='delay in seconds before triggering the failure injection via Gibbon')
    parser.add_argument('--recoveryAction', '-ra', action='store_true', help='flag that indicates if recovery action was executed')
    parser.add_argument('--recoveryDelay', '-rd',type=int,nargs='?',const=0, help='delay in seconds before triggering the recovery action')
    args = parser.parse_args()

    inputPath = args.runFolder[0]
    scenario = args.scenario[0]
    
    

    if scenario == 'elasticity':
        print("creating elasticity phase plots...")
        scaleDelay = args.scaleDelay

        deployVmDuration, scaleOutDuration = getPhaseDurations(inputPath)

        scaleOutStart = scaleDelay + deployVmDuration
        scaleOutEnd = scaleOutStart + scaleOutDuration

        plot_throughput_timeseries(inputPath,scaleOutStart,scaleOutEnd)

    elif scenario == 'availability':
        print("creating availability phase plots...")   
        gibbonDelay = args.gibbonDelay
        recoveryAction = args.recoveryAction
        recoveryDelay = args.recoveryDelay

        if recoveryAction:
            print("plotting availability phases with recovery!")
            deployVmDuration, scaleOutDuration = getPhaseDurations(inputPath)
            
            gibbonDuration = getGibbonPhaseDuration(inputPath)
            gibbonEnd = gibbonDelay + gibbonDuration

            recoveryStart = gibbonEnd + recoveryDelay
            
            recoveryEnd = recoveryStart + deployVmDuration + scaleOutDuration
            
            plot_availability_recovery_throughput_timeseries(inputPath, gibbonDelay, gibbonEnd, recoveryStart, recoveryEnd)

        else:
            print("plotting availability phases without recovery!")
            gibbonDuration = getGibbonPhaseDuration(inputPath)
            gibbonEnd = gibbonDelay + gibbonDuration

            plot_availability_no_recovery_throughput_timeseries(inputPath, gibbonDelay, gibbonEnd)


    print("finished evaluation metrics plotting!")



def plot_throughput_timeseries(inputPath, scaleOutStart, scaleOutEnd):

    dataPath = inputPath + plotting_config.DATA_FOLDER + plotting_config.DATA_TRANSACTION_FILE

    extractedTimeseriesData = TimeseriesData()

    file = open(dataPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                operationItem = splittedLine[2]
                operationItem = operationItem.strip()
                #print (operationItem)
                extractedTimeseriesData.operations.append(float(operationItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")      


    initialAvg = calculate_phase_average(extractedTimeseriesData,0,scaleOutStart)
    print("Initial (overloaded) phase average is: " + str(initialAvg) + " ops/s")
    adaptationAvg = calculate_phase_average(extractedTimeseriesData,scaleOutStart,scaleOutEnd)
    print("Scale-out (overloaded) phase average is: " + str(adaptationAvg) + " ops/s")
    evaluationEnd = max(extractedTimeseriesData.timestamps)
    endAvg = calculate_phase_average(extractedTimeseriesData,scaleOutEnd,evaluationEnd)
    print("After scale-out (overloaded) phase average is: " + str(endAvg) + " ops/s")

    print("plotting throughput timeseries...")
    #Plotting 
    
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Operations / second")
    plt.title("Throughput")

    # Set axes range
    axes = plt.gca()    
    yMax = max(extractedTimeseriesData.operations) + 1000
    axes.set_ylim([0,yMax])
        
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.operations)   

    plt.hlines(y=initialAvg, xmin=0, xmax=scaleOutStart, color='red', linestyles='dashed', zorder=3, label='pre-scale-out avg.')

    plt.hlines(y=adaptationAvg, xmin=scaleOutStart, xmax=scaleOutEnd, color='orange', linestyles='dashed', zorder=3, label='scale-out avg.')

    plt.hlines(y=endAvg, xmin=scaleOutEnd, xmax=evaluationEnd, color='green', linestyles='dashed', zorder=3, label='post-scale-out avg.')

    plt.axvline(scaleOutStart, color='orange', linestyle=':', label="scale-out")

    leg = plt.legend()

    outputFile = inputPath + plotting_config.TIMESERiES_FOLDER + "transaction-throughput-ts-phases.png"

    print("saving throughput plot to disk: " + outputFile)
    plt.savefig(outputFile)   
   

def plot_availability_recovery_throughput_timeseries(inputPath, gibbonStart, gibbonEnd, scaleOutStart, scaleOutEnd):

    dataPath = inputPath + plotting_config.DATA_FOLDER + plotting_config.DATA_TRANSACTION_FILE

    extractedTimeseriesData = TimeseriesData()

    file = open(dataPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                operationItem = splittedLine[2]
                operationItem = operationItem.strip()
                #print (operationItem)
                extractedTimeseriesData.operations.append(float(operationItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")      


    initialAvg = calculate_phase_average(extractedTimeseriesData,0,gibbonStart)
    print("Initial (healthy) phase average is: " + str(initialAvg) + " ops/s")

    unhealthyAvg = calculate_phase_average(extractedTimeseriesData,gibbonEnd,scaleOutStart)
    print("Unhealty phase average is: " + str(unhealthyAvg) + " ops/s")

    recoveryAvg = calculate_phase_average(extractedTimeseriesData,scaleOutStart,scaleOutEnd)
    print("Recovery (unhealthy) phase average is: " + str(recoveryAvg) + " ops/s")

    evaluationEnd = max(extractedTimeseriesData.timestamps)
    healthyAvg = calculate_phase_average(extractedTimeseriesData,scaleOutEnd,evaluationEnd)
    print("After recovery (healthy) phase average is: " + str(healthyAvg) + " ops/s")



    print("plotting throughput timeseries...")
    #Plotting 
    
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Operations / second")
    plt.title("Throughput")

    # Set axes range
    axes = plt.gca()    
    yMax = max(extractedTimeseriesData.operations) + 1000
    axes.set_ylim([0,yMax])


        
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.operations)   

    plt.hlines(y=initialAvg, xmin=0, xmax=gibbonEnd, color='green', linestyles='dashed', zorder=3, label='healthy avg.')

    plt.hlines(y=unhealthyAvg, xmin=gibbonEnd, xmax=scaleOutStart, color='red', linestyles='dashed', zorder=3, label='unhealthy avg.')

    plt.hlines(y=recoveryAvg, xmin=scaleOutStart, xmax=scaleOutEnd, color='orange', linestyles='dashed', zorder=3, label='recovery avg.')

    plt.hlines(y=healthyAvg, xmin=scaleOutEnd, xmax=evaluationEnd, color='green', linestyles='dashed', zorder=3)
    
    plt.axvline(gibbonStart, color='red', linestyle=':', label="failure injection")

    plt.axvline(scaleOutStart, color='orange', linestyle=':', label="recovery start")

    leg = plt.legend()

    outputFile = inputPath + plotting_config.TIMESERiES_FOLDER + "transaction-throughput-ts-phases.png"

    print("saving throughput plot to disk: " + outputFile)
    plt.savefig(outputFile)   


def plot_availability_no_recovery_throughput_timeseries(inputPath, gibbonStart, gibbonEnd):

    dataPath = inputPath + plotting_config.DATA_FOLDER + plotting_config.DATA_TRANSACTION_FILE

    extractedTimeseriesData = TimeseriesData()

    file = open(dataPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                operationItem = splittedLine[2]
                operationItem = operationItem.strip()
                #print (operationItem)
                extractedTimeseriesData.operations.append(float(operationItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")      


    initialAvg = calculate_phase_average(extractedTimeseriesData,0,gibbonStart)
    print("Initial (healthy) phase average is: " + str(initialAvg) + " ops/s")

    evaluationEnd = max(extractedTimeseriesData.timestamps)
    unhealthyAvg = calculate_phase_average(extractedTimeseriesData,gibbonEnd,evaluationEnd)
    print("Unhealty phase average is: " + str(unhealthyAvg) + " ops/s")

   
    print("plotting throughput timeseries...")
    #Plotting 
    
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Operations / second")
    plt.title("Throughput")

    # Set axes range
    axes = plt.gca()    
    yMax = max(extractedTimeseriesData.operations) + 1000
    axes.set_ylim([0,yMax])


        
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.operations)   

    plt.hlines(y=initialAvg, xmin=0, xmax=gibbonEnd, color='green', linestyles='dashed', zorder=3, label='healthy avg.')

    plt.hlines(y=unhealthyAvg, xmin=gibbonEnd, xmax=evaluationEnd, color='red', linestyles='dashed', zorder=3, label='unhealthy avg.')
    
    plt.axvline(gibbonStart, color='red', linestyle=':', label="failure injection")

    leg = plt.legend()

    outputFile = inputPath + plotting_config.TIMESERiES_FOLDER + "transaction-throughput-ts-phases.png"

    print("saving throughput plot to disk: " + outputFile)
    plt.savefig(outputFile)   
           
       
    


def plot_latency_timeseries(inputPath, outputPath):

  
    extractedTimeseriesData = TimeseriesData()
    extractedTimeseriesData.timestamps = []


    file = open(inputPath, 'r', encoding='utf8')
    try:
        for line in file:
            line = line.rstrip()
            if "[TIMESERIES]" in line:
                #extract current ops per second
                splittedLine = line.split(";")
                latencyItem = splittedLine[3]
                latencyItem = latencyItem.strip() 
                #print (latencyItem)
                extractedTimeseriesData.latencies.append(float(latencyItem))
                #extract timestamp
                secondsItem = splittedLine[1]
                secondsItem = secondsItem.strip()
                extractedTimeseriesData.timestamps.append(float(secondsItem))

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")       

    print("plotting latency timeseries...")
    #Plotting 
    plt.figure()
    plt.xlabel("Runtime in sec")
    plt.ylabel("Avg. Latency in us")
    plt.title("Latency")
    plt.plot(extractedTimeseriesData.timestamps, extractedTimeseriesData.latencies)    

    outputFile = outputPath + "-latency-ts.png"
    print("saving latency plot to disk: " + outputFile)
    plt.savefig(outputFile)
    
    
def calculate_phase_average(extractedTimeseriesData, start,end):


    df = pd.DataFrame({'timestamps': extractedTimeseriesData.timestamps,'operations' : extractedTimeseriesData.operations})
    df = df[['timestamps','operations']]

    #print(df.to_string())

    #maximum = df['operations'].max()
    #print('Max value is: ' , maximum)
    #print("phase start: " + str(start) + " phase end: " + str(end))

    if(start > end):
        print("Start of current phase is after end of phase! Did the adapation finish after the transaction phase?")

    phaseDataFrame = df[df['timestamps'].between(start, end, inclusive=True)]
    #print(phaseDataFrame.to_string())

    average = phaseDataFrame['operations'].mean()
    #print('Phase average is: ' , average)

    return average


def getPhaseDurations(inputPath):

    #TODO: currently only supports one adaptation phase by using the second occurrence of the respective tasks!  

    inputPath = inputPath + plotting_config.TASK_FOLDER + plotting_config.TASK_LOG_FILE



    file = open(inputPath, 'r', encoding='utf8')

    deployVmStart = None
    deployVmFinish = None

    deployScaledDbStart = None
    deployScaledDbFinish = None

    try:
        for line in file:
            line = line.rstrip()
            if "SPAWN_VM" in line and "STARTED" in line:
                splittedLine = line.split(";")
                deployVmStart = splittedLine[1]
                deployVmStart = deployVmStart.replace('s','')
                deployVmStart = int(deployVmStart)
            if "SPAWN_VM" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                deployVmFinish = splittedLine[1]
                deployVmFinish = deployVmFinish.replace('s','')
                deployVmFinish = int(deployVmFinish) 
            if "DEPLOY_DB" in line and "STARTED" in line:
                splittedLine = line.split(";")
                deployScaledDbStart = splittedLine[1]
                deployScaledDbStart = deployScaledDbStart.replace('s','')
                deployScaledDbStart = int(deployScaledDbStart)
            if "DEPLOY_DB" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                deployScaledDbFinish = splittedLine[1]
                deployScaledDbFinish = deployScaledDbFinish.replace('s','')
                deployScaledDbFinish = int(deployScaledDbFinish)           

    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")

      

    deployVmDuration = deployVmFinish - deployVmStart
    scaleOutDuration = deployScaledDbFinish - deployScaledDbStart 

    print("VM deploy duration is: " + str(deployVmDuration))
    print("scale-out  duration is: " + str(scaleOutDuration))  

    return deployVmDuration, scaleOutDuration

def getGibbonPhaseDuration(inputPath):

    #TODO: currently only supports one adaptation phase by using the second occurrence of the respective tasks!  

    inputPath = inputPath + plotting_config.TASK_FOLDER + plotting_config.TASK_LOG_FILE

    file = open(inputPath, 'r', encoding='utf8')

    gibbonStart = None
    gibbonFinish = None

    try:
        for line in file:
            line = line.rstrip()
            if "GIBBON" in line and "STARTED" in line:
                splittedLine = line.split(";")
                gibbonStart = splittedLine[1]
                gibbonStart = gibbonStart.replace('s','')
                gibbonStart = int(gibbonStart)
            if "GIBBON" in line and "FINISHED" in line:
                splittedLine = line.split(";")
                gibbonFinish = splittedLine[1]
                gibbonFinish = gibbonFinish.replace('s','')
                gibbonFinish = int(gibbonFinish)  

    
    except UnicodeDecodeError: 
        print("Skipping line! Unable to decode line due to UnicodeDecodeError!")

    gibbonDuration = gibbonFinish - gibbonStart   

    print("Gibbon  duration is: " + str(gibbonDuration))           

    return gibbonDuration

#init function call
plot_timeseries()