package de.uulm.omi.evaluation.scenario;

import static org.jeasy.flows.engine.WorkFlowEngineBuilder.aNewWorkFlowEngine;
import static org.jeasy.flows.workflow.ParallelFlow.Builder.aNewParallelFlow;
import static org.jeasy.flows.workflow.RepeatFlow.Builder.aNewRepeatFlow;
import static org.jeasy.flows.workflow.SequentialFlow.Builder.aNewSequentialFlow;

import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ElasticityYCSBWrite;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadPhaseContext;
import de.uulm.omi.evaluation.task.CheckDeployment;
import de.uulm.omi.evaluation.task.CheckWorkload;
import de.uulm.omi.evaluation.task.CleanUpResources;
import de.uulm.omi.evaluation.task.CleanUpRun;
import de.uulm.omi.evaluation.task.Delay;
import de.uulm.omi.evaluation.task.DeployDB;
import de.uulm.omi.evaluation.task.FetchYcsbResult;
import de.uulm.omi.evaluation.task.InitEvaluation;
import de.uulm.omi.evaluation.task.InitRun;
import de.uulm.omi.evaluation.task.LogTask;
import de.uulm.omi.evaluation.task.Plotting;
import de.uulm.omi.evaluation.task.Plotting.PlottingTask;
import de.uulm.omi.evaluation.task.Scale;
import de.uulm.omi.evaluation.task.TaskState;
import de.uulm.omi.evaluation.task.TaskType;
import de.uulm.omi.evaluation.task.YcsbWorkload;
import de.uulm.omi.utils.YcsbWorkloadType;
import org.jeasy.flows.engine.WorkFlowEngine;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.workflow.SequentialFlow;
import org.jeasy.flows.workflow.WorkFlow;

/**
 * Created by Daniel Seybold on 15.10.2018.
 */
public class ElasticityYCSBWriteScenario extends EvaluationScenario {
  private final ElasticityYCSBWrite elasticityYCSBWrite;
  private final boolean cleanUp;
  private final ClusterContext dbmsClusterContext;
  private final WorkloadPhaseContext workloadPhaseContext;
  private final int overallRuns;

  private int elasticityCounter = 0;

  public ElasticityYCSBWriteScenario(String scenarioType, String cloud, String dbmsType,  String dbmsConfig, Integer runs,
      ElasticityYCSBWrite elasticityYCSBWrite,
      boolean cleanUp) {
    super(scenarioType, cloud, dbmsType,  dbmsConfig, runs);

    //TODO: integrate workloadApiEndpoints for multi-phase workloads

    this.cleanUp = cleanUp;
    this.elasticityYCSBWrite = elasticityYCSBWrite;
    this.dbmsClusterContext = new ClusterContext();
    this.workloadPhaseContext = new WorkloadPhaseContext();
    this.overallRuns = runs;
  }

  @Override
  public WorkReport execute() {

    /**
     **************************** Deployment *********************************************
     */

    /**
     * init general result folders and counters
     */
    InitEvaluation initEvaluationTask = new InitEvaluation(this.evaluationContext);

    /**
     * init evaluation result folder for each evaluation run
     */
    InitRun initRun = new InitRun(this.evaluationContext);

    /**
     * deploy the DBMS cluster
     */
    DeployDB deployDB = new DeployDB(this.evaluationContext, this.elasticityYCSBWrite.getDbmsCluster(),
        this.elasticityYCSBWrite.getDbmsNetwork().getType().name(), this.dbmsClusterContext);


    /**
     * check the deployment process
     */
    CheckDeployment checkDeployment = new CheckDeployment(this.evaluationContext,
        this.dbmsClusterContext);

    /**
     **************************** Workload *********************************************
     */

    /**
     * start the YCSB load workload
     */
    YcsbWorkload ycsbWorkload = new YcsbWorkload(this.evaluationContext, this.dbmsClusterContext,
        this.elasticityYCSBWrite.getWorkloadLoad(), this.elasticityYCSBWrite.getWorkloadNetwork(), this.workloadPhaseContext,
        YcsbWorkloadType.LOAD,this.elasticityYCSBWrite.getWorkloadInstances());

    /**
     * check the state of the running workload
     */
    CheckWorkload checkWorkload = new CheckWorkload(this.evaluationContext, this.workloadPhaseContext,this.elasticityYCSBWrite
        .getWorkloadInstances());

    /**
     **************************** Elasticity *********************************************
     * all elasticity task are specified directly within the flows as multiple steps are executed
     */



    /**
     **************************** Plotting *********************************************
     */

    /**
     * Fetch the workload results for the current run and store to disk
     */
    FetchYcsbResult fetchYcsbResult = new FetchYcsbResult(this.evaluationContext,this.workloadPhaseContext, YcsbWorkloadType.LOAD,this.elasticityYCSBWrite
        .getWorkloadInstances());

    /**
     * CleanUpResources all resources of the current run
     */
    CleanUpResources cleanUpResources = new CleanUpResources(this.evaluationContext, this.dbmsClusterContext, this.cleanUp);

    /**
     * Plot the timeseries of the current run
     */
    Plotting mergeYCSB = new Plotting(this.evaluationContext, this.dbmsClusterContext, PlottingTask.MERGE_YCSB_LOAD);

    Plotting evaluationTimeseries = new Plotting(this.evaluationContext, this.dbmsClusterContext, PlottingTask.EVALUATION_METRICS);

    Plotting systemTimeseries = new Plotting(this.evaluationContext, this.dbmsClusterContext, PlottingTask.SYSTEM_METRICS);

    Plotting workloadTimeseries = new Plotting(this.evaluationContext, this.dbmsClusterContext, PlottingTask.WORKLOAD_METRICS, this.elasticityYCSBWrite
        .getWorkloadInstances());

    Plotting elasticityPhases = new Plotting(this.evaluationContext, this.dbmsClusterContext, PlottingTask.ELASTICITY_PHASES, this.elasticityYCSBWrite
        .getElasticitySpec().get(0).getScalingDelay().intValue());

    /**
     * Increase the run counter
     */
    CleanUpRun cleanUpRun = new CleanUpRun(this.evaluationContext, this.dbmsClusterContext);

    /**
     * Build the workflow with the created tasks
     */
    SequentialFlow runFlow;


    runFlow = aNewSequentialFlow()
        .execute(initRun)
        .execute(deployDB)
        .execute(new LogTask(evaluationContext, TaskType.DEPLOY_DB, TaskState.STARTED))
        .execute(checkDeployment)
        .execute(new LogTask(evaluationContext, TaskType.DEPLOY_DB, TaskState.FINISHED))
        .execute(
            aNewParallelFlow()
                .execute(
                    aNewSequentialFlow()
                        .execute(ycsbWorkload)
                        .execute(new LogTask(evaluationContext, TaskType.YCSB_LOAD, TaskState.STARTED))
                        .execute(checkWorkload)
                        .execute(new LogTask(evaluationContext, TaskType.YCSB_LOAD, TaskState.FINISHED))
                        .build(),
                    //aNewRepeatFlow()
                      //  .repeat(
                            aNewSequentialFlow()
                                .execute(new Delay(evaluationContext, this.elasticityYCSBWrite.getElasticitySpec().get(this.elasticityCounter).getScalingDelay().intValue()))
                                .execute(new LogTask(evaluationContext, TaskType.SCALE, TaskState.STARTED))
                                .execute(new Scale(evaluationContext, this.dbmsClusterContext, true, this.elasticityYCSBWrite
                                    .getElasticitySpec()))
                                .execute(new LogTask(evaluationContext, TaskType.SCALE, TaskState.FINISHED))
                                //.execute(new IncreaseCounter(evaluationContext, this.elasticityCounter))
                                .build()
                        //)
                        //.times(this.elasticityCounter)
                        //.build()
                )
                .build()
        )
        .execute(fetchYcsbResult)
        .execute(mergeYCSB)
        .execute(evaluationTimeseries)
        .execute(workloadTimeseries)
        .execute(systemTimeseries)
        .execute(elasticityPhases)
        .execute(new LogTask(evaluationContext, TaskType.CLEANUP, TaskState.STARTED))
        .execute(cleanUpResources)
        .execute(new LogTask(evaluationContext, TaskType.CLEANUP, TaskState.FINISHED))
        .execute(cleanUpRun)
        .build();





    /**
     * execute the specified workflows
     */
    WorkFlow sensorStorageWorkFlow = aNewSequentialFlow()
        .execute(initEvaluationTask)
        .then(
            aNewRepeatFlow()
                .repeat(
                    runFlow
                )
                .times(this.overallRuns)
                .build()
        )
        .build();

    WorkFlowEngine workFlowEngine = aNewWorkFlowEngine().build();
    WorkReport workReport = workFlowEngine.run(sensorStorageWorkFlow);

    return  workReport;

  }
}
