package de.uulm.omi.evaluation.task;

import static com.google.common.base.Preconditions.checkNotNull;

import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.utils.ColosseumUtils;
import de.uulm.omi.utils.DBMSNodeType;
import de.uulm.omi.utils.EntrypointProperties;
import java.io.File;
import java.io.IOException;
import java.lang.ProcessBuilder.Redirect;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 04.07.2018.
 */
public class Plotting extends EvaluationTask {

  private String loggerPrefix;
  private ClusterContext clusterContext;
  private PlottingTask plottingTask;
  private List<String> workloadInstances;
  private int scalingDelay;
  private int gibbonDelay;

  public enum PlottingTask{
    SYSTEM_METRICS,
    EVALUATION_METRICS,
    WORKLOAD_METRICS,
    AGGREGATE_EVALUATION_METRICS,
    MERGE_YCSB_LOAD,
    MERGE_YCSB_TRANSACTION,
    MERGE_YCSB_ALL,
    ELASTICITY_PHASES,
    AVAILABILITY_PHASES,
    AVAILABILITY_PHASES_RECOVERY

  }

  /**
   * general plotting
   * @param evaluationContext
   * @param clusterContext
   * @param plottingTask
   */
  public Plotting(EvaluationContext evaluationContext, ClusterContext clusterContext, PlottingTask plottingTask) {
    super(TaskType.PLOTTING, evaluationContext);



    this.clusterContext = clusterContext;
    this.plottingTask = plottingTask;
  }

  /**
   * System metrics
   * @param evaluationContext
   * @param clusterContext
   * @param plottingTask
   * @param workloadInstances
   */
  public Plotting(EvaluationContext evaluationContext, ClusterContext clusterContext, PlottingTask plottingTask, List<String> workloadInstances) {
    super(TaskType.PLOTTING, evaluationContext);

    this.workloadInstances = workloadInstances;

    this.clusterContext = clusterContext;
    this.plottingTask = plottingTask;
  }

  /**
   * Elasticity Phases
   * @param evaluationContext
   * @param clusterContext
   * @param plottingTask
   * @param scalingDelay
   */
  public Plotting(EvaluationContext evaluationContext, ClusterContext clusterContext, PlottingTask plottingTask, int scalingDelay) {
    super(TaskType.PLOTTING, evaluationContext);

    this.scalingDelay = scalingDelay;

    this.clusterContext = clusterContext;
    this.plottingTask = plottingTask;
  }

  /**
   * Availability Phases
   * @param evaluationContext
   * @param clusterContext
   * @param plottingTask
   * @param gibbonDelay
   * @param scalingDelay
   */
  public Plotting(EvaluationContext evaluationContext, ClusterContext clusterContext, PlottingTask plottingTask, int gibbonDelay, int scalingDelay) {
    super(TaskType.PLOTTING, evaluationContext);

    this.gibbonDelay = gibbonDelay;
    this.scalingDelay = scalingDelay;

    this.clusterContext = clusterContext;
    this.plottingTask = plottingTask;
  }

  @Override
  public WorkReport call() {

    this.loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";


    this.evaluationContext.setEndTime(System.currentTimeMillis()/1000);
    LOGGER.debug(this.loggerPrefix + " evaluation run end time: " + this.evaluationContext.getEndTime());

    switch (this.plottingTask){
      case EVALUATION_METRICS:
        LOGGER.debug(loggerPrefix + "Plotting evaluation metrics...");
        this.plotEvaluationMetrics("load");
        this.plotEvaluationMetrics("transaction");
        break;
      case SYSTEM_METRICS:
        LOGGER.debug(loggerPrefix + "Plotting system metrics...");
        this.plotSystemMetrics();
        break;
      case WORKLOAD_METRICS:
        LOGGER.debug(loggerPrefix + "Plotting workload-API system metrics...");
        this.plotWorkloadApiMetrics();
        break;
      case AGGREGATE_EVALUATION_METRICS:
        LOGGER.debug(loggerPrefix + "Aggregating evaluation metrics...");
        this.aggregateEvaluationMetrics();
        break;
      case MERGE_YCSB_LOAD:
        LOGGER.debug(loggerPrefix + "merging YCSB load metrics...");
        this.mergeYCSBMetrics("load");
        break;
      case MERGE_YCSB_TRANSACTION:
        LOGGER.debug(loggerPrefix + "merging YCSB transaction metrics...");
        this.mergeYCSBMetrics("transaction");
        break;
      case MERGE_YCSB_ALL:
        LOGGER.debug(loggerPrefix + "merging YCSB load and transaction metrics...");
        this.mergeYCSBMetrics("all");
        break;
      case ELASTICITY_PHASES:
        LOGGER.debug(loggerPrefix + "plotting elasticity phases of YCSB results...");
        this.plotElasticityPhases(this.scalingDelay);
        break;
      case AVAILABILITY_PHASES:
        LOGGER.debug(loggerPrefix + "plotting availability phases of YCSB results...");
        this.plotAvailabilityPhases(this.gibbonDelay);
        break;
      case AVAILABILITY_PHASES_RECOVERY:
        LOGGER.debug(loggerPrefix + "plotting availability phases with recovery of YCSB results...");
        this.plotElasticityPhases(this.scalingDelay);
        break;

    }

    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }


  private void mergeYCSBMetrics(String resultType){

    LOGGER.debug(this.loggerPrefix + "merging YCSB evaluation metrics!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "utils" + File.separator + "mergeYCSB.py";

    String mergeResultFolder = this.evaluationContext.getEvluationRun() + File.separator
        + EvaluationTask.DATA_FOLDER + File.separator;


    ProcessBuilder mergeYCSB = new ProcessBuilder(
        "python3",
        plottingScript,
        "-rf",
        mergeResultFolder,
        "-wp",
        resultType
    );

    LOGGER.debug(this.loggerPrefix + "Calling aggregation script: " + mergeYCSB.command().toString());



    try {

      File mergeYCSBOutput = new File(mergeResultFolder + File.separator + "merge.log");
      mergeYCSB.redirectErrorStream(true);
      mergeYCSB.redirectOutput(Redirect.appendTo(mergeYCSBOutput));

      mergeYCSB.start();

    } catch (IOException e) {
      LOGGER.error(this.loggerPrefix + "Error while aggregating evaluation metrics!",e);
    }



  }

  private void plotEvaluationMetrics(String resultType){

    LOGGER.debug(this.loggerPrefix + "plotting evaluation timeseries!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "timeseries" + File.separator + "plotEvaluationMetrics.py";

    String dataDir = this.evaluationContext.getEvluationRun()
       + File.separator + EvaluationTask.DATA_FOLDER;

    String timeseriesDir = this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.TIMESERIES_FOLDER;

    ProcessBuilder timeseriesPlottingProcess = new ProcessBuilder(
        "python3",
        plottingScript,
        "-i",
        dataDir + File.separator +  resultType + ".txt",
        "-o",
        timeseriesDir + File.separator +  resultType);

    LOGGER.debug(this.loggerPrefix + "Calling plotting script: " + timeseriesPlottingProcess.command().toString());

    try {

      File plottingOutput = new File(timeseriesDir + File.separator +  resultType + "-ts-log.txt");
      timeseriesPlottingProcess.redirectErrorStream(true);
      timeseriesPlottingProcess.redirectOutput(Redirect.appendTo(plottingOutput));

      timeseriesPlottingProcess.start();

    } catch (IOException e) {
      LOGGER.error(this.loggerPrefix + "Error while plotting " + resultType + " evaluation timeseries!",e);
    }


  }

  private List<String> getVmIps(){

    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());
    checkNotNull(appInstanceId);

    //build endpoints
    List<String> endpoints = ColosseumUtils.getEndpoints(appInstanceId, "PUBLIC", DBMSNodeType.ALL);
    if(endpoints.size() == 0){
      LOGGER.warn(loggerPrefix + "Found no endpoints for DBMS with ApplicationInstanceId: " + this.clusterContext.getIdApplicationInstance()  +" ! Aborting!");
    }

    return endpoints;

  }

  private void plotWorkloadApiMetrics() {

    LOGGER.debug(this.loggerPrefix + "plotting Workload-API system metric timeseries!");

    String plottingScript =
        this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator
            + "timeseries" + File.separator + "plotSystemMetrics.py";

    String monitoringDir = this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.MONITORING_FOLDER;

    for (String workloadApiEndpoint : this.workloadInstances) {

      String IPADDRESS_PATTERN =
          "(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)";

      Pattern pattern = Pattern.compile(IPADDRESS_PATTERN);
      Matcher matcher = pattern.matcher(workloadApiEndpoint);
      String workloadApiIp = "";
      if (matcher.find()) {
        workloadApiIp = matcher.group();
      } else {
        LOGGER.warn("No WorkloadAPI IP address could be found in: " + workloadApiEndpoint);
      }

      ProcessBuilder systemMetricPlottingProcess = new ProcessBuilder(
          "python3",
          plottingScript,
          "--influxDB",
          this.properties.get(EntrypointProperties.INFLUXDB_DATABASE).toString(),
          "--influxUrl",
          this.properties.get(EntrypointProperties.INFLUXDB_URL).toString(),
          "--outputFolder",
          monitoringDir,
          "--startTime",
          String.valueOf(this.evaluationContext.getStartTime()),
          "--endTime",
          String.valueOf(this.evaluationContext.getEndTime()),
          "--metric",
          "all",
          "--publicIpTag",
          workloadApiIp,
          "--nodeTypeTag",
          "WorkloadInstance");

      LOGGER.debug(
          this.loggerPrefix + "Calling plotting script: " + systemMetricPlottingProcess.command()
              .toString());

      try {

        File plottingOutput = new File(
            monitoringDir + File.separator + "node-workload-" + workloadApiIp + ".log");
        systemMetricPlottingProcess.redirectErrorStream(true);
        systemMetricPlottingProcess.redirectOutput(Redirect.appendTo(plottingOutput));

        systemMetricPlottingProcess.start();

      } catch (IOException e) {
        LOGGER.error(
            this.loggerPrefix + "Error while plotting workload API system metric timeseries!", e);
      }
    }


  }

  private void plotSystemMetrics(){

    List<String> publicIpEndpoints = this.getVmIps();

    LOGGER.debug(this.loggerPrefix + "plotting system metric timeseries!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "timeseries" + File.separator + "plotSystemMetrics.py";

    String monitoringDir = this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.MONITORING_FOLDER;

    for (String publicIp : publicIpEndpoints) {

      ProcessBuilder systemMetricPlottingProcess = new ProcessBuilder(
          "python3",
          plottingScript,
          "--influxDB",
          this.properties.get(EntrypointProperties.INFLUXDB_DATABASE).toString(),
          "--influxUrl",
          this.properties.get(EntrypointProperties.INFLUXDB_URL).toString(),
          "--outputFolder",
          monitoringDir,
          "--startTime",
          String.valueOf(this.evaluationContext.getStartTime()),
          "--endTime",
          String.valueOf(this.evaluationContext.getEndTime()),
          "--metric",
          "all",
          "--publicIpTag",
          publicIp,
          "--nodeTypeTag",
          "DBMS");

      LOGGER.debug(this.loggerPrefix + "Calling plotting script: " + systemMetricPlottingProcess.command().toString());



      try {

        File plottingOutput = new File(monitoringDir + File.separator + "node-" + publicIp + ".log");
        systemMetricPlottingProcess.redirectErrorStream(true);
        systemMetricPlottingProcess.redirectOutput(Redirect.appendTo(plottingOutput));

        systemMetricPlottingProcess.start();

      } catch (IOException e) {
        LOGGER.error(this.loggerPrefix + "Error while plotting system metric timeseries!",e);
      }

    }


  }

  private void aggregateEvaluationMetrics(){

    LOGGER.debug(this.loggerPrefix + "aggregating evaluation metrics!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "boxplots" + File.separator + "generatePlotData.py";

    String aggregationResultFolder = this.properties.getProperty(EntrypointProperties.SCENARIO_DIR) + File.separator + this.evaluationContext.getScenarioType() + File.separator
        + this.evaluationContext.getCloudType() + File.separator
        + this.evaluationContext.getDbmsType() + File.separator + this.evaluationContext.getDbmsConfiguration();


      ProcessBuilder aggregateEvaluationMetrics = new ProcessBuilder(
          "python3",
          plottingScript,
          "-rf",
          aggregationResultFolder,
          "-wp",
          "load"
          );

      LOGGER.debug(this.loggerPrefix + "Calling aggregation script: " + aggregateEvaluationMetrics.command().toString());



      try {

        File aggregationOutput = new File(aggregationResultFolder + File.separator + "plots" + File.separator + "aggregation.log");
        aggregateEvaluationMetrics.redirectErrorStream(true);
        aggregateEvaluationMetrics.redirectOutput(Redirect.appendTo(aggregationOutput));

        aggregateEvaluationMetrics.start();

      } catch (IOException e) {
        LOGGER.error(this.loggerPrefix + "Error while aggregating evaluation metrics!",e);
      }

  }


  private void plotElasticityPhases(int scaleOutDelay){
    LOGGER.debug(this.loggerPrefix + "plotting elasticity phases!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "timeseries" + File.separator + "plotEvaluationPhases.py";

    String resultFolder = this.evaluationContext.getEvluationRun() + File.separator;

    ProcessBuilder plotElasticityPhases = new ProcessBuilder(
        "python3",
        plottingScript,
        "-rf",
        resultFolder,
        "-s",
        "elasticity",
        "-sd",
        String.valueOf(scaleOutDelay)
    );

    LOGGER.debug(this.loggerPrefix + "Calling elasticity phase plotting script: " + plotElasticityPhases.command().toString());



    try {

      File plotElasticityPhasesOutput = new File(resultFolder + File.separator + "timeseries" + File.separator + "elasticityPhases.log");
      plotElasticityPhases.redirectErrorStream(true);
      plotElasticityPhases.redirectOutput(Redirect.appendTo(plotElasticityPhasesOutput));

      plotElasticityPhases.start();

    } catch (IOException e) {
      LOGGER.error(this.loggerPrefix + "Error while plotting elasticity phases!",e);
    }

  }

  private void plotAvailabilityPhases(int gibbonDelay){
    LOGGER.debug(this.loggerPrefix + "plotting availability phases!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "timeseries" + File.separator + "plotEvaluationPhases.py";

    String resultFolder = this.evaluationContext.getEvluationRun() + File.separator;

    ProcessBuilder plotAvailabilityPhases = new ProcessBuilder(
        "python3",
        plottingScript,
        "-rf",
        resultFolder,
        "-s",
        "availability",
        "-gd",
        String.valueOf(gibbonDelay)
    );

    LOGGER.debug(this.loggerPrefix + "Calling availability phase plotting script: " + plotAvailabilityPhases.command().toString());



    try {

      File plotAvailabilityPhasesOutput = new File(resultFolder + File.separator + "timeseries" + File.separator + "availabilityPhases.log");
      plotAvailabilityPhases.redirectErrorStream(true);
      plotAvailabilityPhases.redirectOutput(Redirect.appendTo(plotAvailabilityPhasesOutput));

      plotAvailabilityPhases.start();

    } catch (IOException e) {
      LOGGER.error(this.loggerPrefix + "Error while plotting availability phases!",e);
    }

  }

  private void plotAvailabilityRecoveryPhases(int gibbonDelay, int scaleDelay){
    LOGGER.debug(this.loggerPrefix + "plotting availability with recovery phases!");

    String plottingScript = this.properties.get(EntrypointProperties.PLOTTING_SCRIPT_DIR) + File.separator + "timeseries" + File.separator + "plotEvaluationPhases.py";

    String resultFolder = this.evaluationContext.getEvluationRun() + File.separator;

    ProcessBuilder plotAvailabilityPhases = new ProcessBuilder(
        "python3",
        plottingScript,
        "-rf",
        resultFolder,
        "-s",
        "availability",
        "-gd",
        String.valueOf(gibbonDelay),
        "-rd",
        String.valueOf(scaleDelay),
        "-ra"
    );

    LOGGER.debug(this.loggerPrefix + "Calling availability with recovery phase plotting script: " + plotAvailabilityPhases.command().toString());



    try {

      File plotAvailabilityPhasesOutput = new File(resultFolder + File.separator + "timeseries" + File.separator + "availabilityPhases.log");
      plotAvailabilityPhases.redirectErrorStream(true);
      plotAvailabilityPhases.redirectOutput(Redirect.appendTo(plotAvailabilityPhasesOutput));

      plotAvailabilityPhases.start();

    } catch (IOException e) {
      LOGGER.error(this.loggerPrefix + "Error while plotting availability with recovery phases!",e);
    }

  }

}
