package de.uulm.omi.evaluation.task;

import static com.google.common.base.Preconditions.checkNotNull;

import com.github.rholder.retry.RetryException;
import com.github.rholder.retry.Retryer;
import com.github.rholder.retry.RetryerBuilder;
import com.github.rholder.retry.StopStrategies;
import com.github.rholder.retry.WaitStrategies;
import com.google.common.base.Predicates;
import de.uniulm.omi.cloudiator.colosseum.client.Client;
import de.uniulm.omi.cloudiator.colosseum.client.entities.ApplicationComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Instance;
import de.uniulm.omi.cloudiator.colosseum.client.entities.LifecycleComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.VirtualMachine;
import de.uniulm.omi.cloudiator.colosseum.client.entities.enums.RemoteState;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ElasticityStep;
import de.uulm.omi.cloudiator.dbms.evaluator.model.ElasticityStep.ScalingTypeEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.utils.ColosseumClientHelper;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 05.03.2018.
 */
public class Scale extends EvaluationTask {

  private final Client colosseumClient;

  private final long VM_CREATION_WAITING_PERIOD = 10000;
  private final long INSTANCE_CREATION_WAITING_PERIOD = 20000;

  private final boolean trigger;
  private final ClusterContext clusterContext;
  private final List<ElasticityStep> elasticitySpec;


  public Scale(EvaluationContext evaluationContext, ClusterContext dbmsClusterContext,
      boolean trigger, List<ElasticityStep> elasticitySpec) {
    super(TaskType.SCALE, evaluationContext);
    this.colosseumClient = ColosseumClientHelper.getColosseumClient();
    this.trigger = trigger;
    this.clusterContext = dbmsClusterContext;
    this.elasticitySpec = elasticitySpec;
  }


  @Override
  public WorkReport call() {

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";

    if(!trigger){
      LOGGER.info(loggerPrefix + "Skipping Scale-out task as specified!");
      return new DefaultWorkReport(WorkStatus.COMPLETED);

    }

    String specUri =  this.evaluationContext.getEvluationRun()
        + File.separator + EvaluationTask.SPECS_FOLDER
        + File.separator + EvaluationTask.ELASTICITY_SPEC;
    this.writeReport(specUri, GSON.toJson(this.elasticitySpec));

    if(elasticitySpec.get(0).getScalingType().equals(ScalingTypeEnum.OUT)){
      LOGGER.info(loggerPrefix + " Executing SCALE-OUT as specified!");
    }else{
      LOGGER.warn(loggerPrefix + " Executing SCALE-IN as specified is currently not implemented, skipping!!");
      return new DefaultWorkReport(WorkStatus.FAILED);
    }



    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());
    checkNotNull(appInstanceId);
    LOGGER.debug("Starting scaling for applicationInstanceId: " + appInstanceId);


    //TODO: add node type as parameter, for now only DATA nodes can be scaled out
    //TODO: implemented scale-in logic

    List<Instance> instances = colosseumClient.controller(Instance.class).getList();
    Instance scaleOutTemplateInstance = null;

    //Search for matching DATA component VM
    for (Instance instance : instances){
      if(instance.getApplicationInstance().equals(appInstanceId)){

        //check for DATA component
        ApplicationComponent applicationComponent = colosseumClient.controller(ApplicationComponent.class).get(instance.getApplicationComponent());

        LifecycleComponent component = colosseumClient.controller(LifecycleComponent.class)
            .get(applicationComponent.getComponent());

        if(component.getName().equals("DATA")){
          scaleOutTemplateInstance = instance;
          break;
        }

      }

    }

    if(scaleOutTemplateInstance == null){
      LOGGER.warn(loggerPrefix + "No matching DATA component found applicationInstanceId: " + appInstanceId + "! Cancelling scale-out!");
      return new DefaultWorkReport(WorkStatus.FAILED);

    }else {
      LOGGER.debug(loggerPrefix + "Found  matching DATA component with applicationInstanceId: " + appInstanceId );
    }

    //scale-out VM
    VirtualMachine scaleOutVm = this.scaleOutVM(scaleOutTemplateInstance);
    LOGGER.debug(loggerPrefix + "Successfully sent create VM request!");



    //scale-out Instance
    Instance scaleOutInstance = this.scaleOutInstance(scaleOutTemplateInstance, scaleOutVm);
    LOGGER.debug(loggerPrefix + "Successfully sent create Instance request!");


    //wait until VM is created
    LOGGER.debug(loggerPrefix + "Waiting for VM creation");
    new LogTask(evaluationContext, TaskType.SPAWN_VM, TaskState.STARTED).call();
    if(!checkIfVmIsReady(scaleOutVm)){
      LOGGER.debug(loggerPrefix + "Create scale-out VM failed!");
      return new DefaultWorkReport(WorkStatus.FAILED);
    }
    new LogTask(evaluationContext, TaskType.SPAWN_VM, TaskState.FINISHED).call();


    //wait until DBMS is deployed on VM
    LOGGER.debug(loggerPrefix + "Waiting for Instance creation!");
    new LogTask(evaluationContext, TaskType.DEPLOY_DB, TaskState.STARTED).call();
    waitUntilInstanceIsReady(scaleOutInstance);
    new LogTask(evaluationContext, TaskType.DEPLOY_DB, TaskState.FINISHED).call();
    /*
    if(!checkIfInstanceIsReady(scaleOutInstance)){
      LOGGER.debug(loggerPrefix + "Create scale-out Instance failed!");
      return new DefaultWorkReport(WorkStatus.FAILED);
    }
    */

    LOGGER.info(loggerPrefix + "Scale-out successful, created new component instance with ID: " + scaleOutInstance.getId());

    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }

  private VirtualMachine scaleOutVM(Instance scaleOutTemplateInstance){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";


    Long vmId = scaleOutTemplateInstance.getVirtualMachine();

    VirtualMachine virtualMachine = colosseumClient.controller(VirtualMachine.class).get(vmId);

    VirtualMachine scaleOutVm = new VirtualMachine(
        null,
        null,
        null,
        null,
        virtualMachine.getCloud(),
        virtualMachine.getCloudCredentials(),
        virtualMachine.getOwner(),
        virtualMachine.getLocation(),
        virtualMachine.getName() + "ScaleOut_" + System.currentTimeMillis(),
        virtualMachine.getImage(),
        virtualMachine.getHardware(),
        virtualMachine.getTemplateOptions()
    );

    LOGGER.debug(loggerPrefix + "Creating new VM for scale-out...");
    scaleOutVm = colosseumClient.controller(VirtualMachine.class).create(scaleOutVm);

    return scaleOutVm;
  }

  private Instance scaleOutInstance(Instance scaleOutTemplateInstance, VirtualMachine scaleOutVm){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";

    Instance scaleOutInstance = new Instance(
        null,
        null,
        null,
        null,
        scaleOutTemplateInstance.getApplicationComponent(),
        scaleOutTemplateInstance.getApplicationInstance(),
        scaleOutVm.getId()
    );

    LOGGER.debug(loggerPrefix + "Creating new DATA instance for scale-out...");
    scaleOutInstance = colosseumClient.controller(Instance.class).create(scaleOutInstance);

    return scaleOutInstance;
  }

  private boolean checkIfVmIsReady(VirtualMachine virtualMachine){
    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";
    while (true){
      VirtualMachine createdVM = colosseumClient.controller(VirtualMachine.class).get(virtualMachine.getId());
      if(createdVM.getRemoteState() != null && createdVM.getRemoteState().equals(RemoteState.ERROR)){
        return false;
      }else if(createdVM.getRemoteState() != null && createdVM.getRemoteState().equals(RemoteState.OK)){
        return true;
      }

      LOGGER.debug(loggerPrefix + "VM creating still in progress, waiting for " + VM_CREATION_WAITING_PERIOD/1000 + " seconds!");
      try {

        Thread.sleep(VM_CREATION_WAITING_PERIOD);

      } catch (InterruptedException e) {
        LOGGER.error(loggerPrefix + "Thread got interrupted while waiting for creating VM...", e);
      }
    }

  }

  private boolean checkIfInstanceIsReady(Instance instance){
    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";
    while (true){
      Instance createdInstance = colosseumClient.controller(Instance.class).get(instance.getId());
      if(createdInstance != null && createdInstance.getRemoteState().equals(RemoteState.ERROR)){
        LOGGER.info(loggerPrefix + "Instance resulted in error state, aborting!");
        return false;
      }else if(createdInstance != null &&createdInstance.getRemoteState().equals(RemoteState.OK)){
        LOGGER.error(loggerPrefix + "Instance successfully created and in OK state!");
        return true;
      }

      LOGGER.debug(loggerPrefix + "Instance creating still in progress, waiting for " + INSTANCE_CREATION_WAITING_PERIOD/1000 + " seconds!");
      try {

        this.wait();
        Thread.sleep(INSTANCE_CREATION_WAITING_PERIOD);
      } catch (InterruptedException e) {
        LOGGER.error(loggerPrefix + "Thread got interrupted while waiting for creating instance...", e);
      }
    }

  }

  private void waitUntilInstanceIsReady(Instance instance){

    String loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +":";

    Callable<Boolean> checkInstanceTask = new Callable<Boolean>() {
      public Boolean call() throws Exception {
        Instance createdInstance = colosseumClient.controller(Instance.class).get(instance.getId());

        if(createdInstance != null && createdInstance.getRemoteState() != null &&  createdInstance.getRemoteState().equals(RemoteState.ERROR)){
          LOGGER.error(loggerPrefix + "Instance resulted in error state, aborting!");
          return false;

        }else if(createdInstance != null && createdInstance.getRemoteState() != null && createdInstance.getRemoteState().equals(RemoteState.OK)){
          LOGGER.info(loggerPrefix + "Instance successfully created and in OK state!");
          return true;
        }

        LOGGER.debug(loggerPrefix + "Instance creating still in progress, waiting for " + INSTANCE_CREATION_WAITING_PERIOD/1000 + " seconds!");
        return false;
      }
    };

    Retryer<Boolean> retryer = RetryerBuilder.<Boolean>newBuilder()
        .retryIfResult(Predicates.<Boolean>equalTo(false))
        .retryIfExceptionOfType(IOException.class)
        .withWaitStrategy(WaitStrategies.fixedWait(10, TimeUnit.SECONDS))
        .withStopStrategy(StopStrategies.neverStop())
        .build();

    try {
      retryer.call(checkInstanceTask);
    } catch (RetryException e) {
      LOGGER.error(loggerPrefix + e.getMessage() );
      e.printStackTrace();
    } catch (ExecutionException e) {
      LOGGER.error(loggerPrefix + e.getMessage() );
      e.printStackTrace();
    }
  }
}
