package de.uulm.omi.evaluation.task;


import static com.google.common.base.Preconditions.checkNotNull;

import de.uulm.omi.cloudiator.dbms.evaluator.model.ClusterContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoints;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DatabaseConfig.DatabaseBindingEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.EvaluationNetwork;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadPhaseContext;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadProcessInstance;
import de.uulm.omi.cloudiator.dbms.evaluator.model.YCSBWorkload;
import de.uulm.omi.dbms.workload.api.DefaultApi;
import de.uulm.omi.utils.ColosseumUtils;
import de.uulm.omi.utils.DBMSNodeType;
import de.uulm.omi.utils.WorkloadApiClientHelper;
import de.uulm.omi.utils.YcsbWorkloadType;
import java.io.File;
import java.util.List;
import org.jeasy.flows.work.DefaultWorkReport;
import org.jeasy.flows.work.WorkReport;
import org.jeasy.flows.work.WorkStatus;

/**
 * Created by Daniel Seybold on 23.01.2018.
 */
public class YcsbWorkload extends EvaluationTask {

  private final YCSBWorkload ycsbWorkload;
  private final EvaluationNetwork network;
  private final ClusterContext clusterContext;
  private final WorkloadPhaseContext workloadPhaseContext;
  private final YcsbWorkloadType ycsbWorkloadType;
  private final List<String> workloadInstances;

  private String loggerPrefix = "";

  public YcsbWorkload(EvaluationContext evaluationContext, ClusterContext clusterContext,
      YCSBWorkload ycsbWorkload, EvaluationNetwork network,
      WorkloadPhaseContext workloadPhaseContext, YcsbWorkloadType ycsbWorkloadType, List<String> workloadInstances) {
    super(TaskType.YCSB_LOAD, evaluationContext);
    this.ycsbWorkload = ycsbWorkload;
    this.network = network;
    this.clusterContext = clusterContext;
    this.workloadPhaseContext = workloadPhaseContext;
    this.ycsbWorkloadType = ycsbWorkloadType;
    this.workloadInstances = workloadInstances;
  }


  @Override
  public WorkReport call() {

    this.loggerPrefix= "RUN " + this.evaluationContext.getRunCounter().intValue() +": ";

    //get workloadApi client
    List<DefaultApi> workloadClients = WorkloadApiClientHelper.getWorkloadClients(workloadInstances);


    Long appInstanceId = Long.valueOf(this.clusterContext.getIdApplicationInstance());
    checkNotNull(appInstanceId);

    //build endpoints
    DBEndpoints dbEndpoints = null;
    if(ycsbWorkload.getDatabaseConfig().getDatabaseBinding().equals(DatabaseBindingEnum.MONGODB)){
      //distinguish between single instance and cluster: if management node available = cluster otherwise single node
      dbEndpoints = ColosseumUtils.extractDBEndpoints(appInstanceId, this.network.getType().toString(), DBMSNodeType.MANAGEMENT);
      if(dbEndpoints.isEmpty()){
        LOGGER.warn(loggerPrefix + "Did  not find any endpoints for MongoDB management node, now looking for MongoDB seed node of a single node cluster!");
        dbEndpoints = ColosseumUtils.extractDBEndpoints(appInstanceId, this.network.getType().toString(), DBMSNodeType.SEED);
      }
    }else {
      dbEndpoints = ColosseumUtils.extractDBEndpoints(appInstanceId, this.network.getType().toString(), DBMSNodeType.ALL);
    }

    if(dbEndpoints.isEmpty()){
      LOGGER.warn(loggerPrefix + "Found no endpoints for applicationInstanceId! Aborting!");
      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    ycsbWorkload.setDbEndpoints(dbEndpoints);






    try {

      String workloadTask = "ycsb_" + evaluationContext.getRunCounter().intValue() + "_" + this.evaluationContext.getScenarioType() + "_" + ycsbWorkloadType.toString();

      int instanceCounter = 0;
      if(this.ycsbWorkloadType.equals(YcsbWorkloadType.LOAD)){


        for (DefaultApi workloadClient: workloadClients) {

          setLoadOperationsPerInstance(instanceCounter);

          writeWorkloadSpec(instanceCounter, LOAD_SPEC);

          WorkloadProcessInstance result = workloadClient.workloadYcsbPost(workloadTask ,ycsbWorkload);

          //TODO: refactor to a list of workloadPhase states
          LOGGER.debug(loggerPrefix + "Status: " + result.getProcessStatus().getValue());
          this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());

          instanceCounter++;
        }

      }else{

        setTransactionOperationsPerInstance();

        for (DefaultApi workloadClient: workloadClients) {

          writeWorkloadSpec(instanceCounter, TRANSACTION_SPEC);

          WorkloadProcessInstance result = workloadClient.workloadYcsbPost(workloadTask ,ycsbWorkload);

          //TODO: refactor to a list of workloadPhase states
          LOGGER.debug(loggerPrefix + "Status: " + result.getProcessStatus().getValue());
          this.workloadPhaseContext.setProcessState(result.getProcessStatus().getValue());

          instanceCounter++;
        }

      }

    } catch (de.uulm.omi.dbms.workload.client.ApiException e) {
      LOGGER.error(loggerPrefix + "error calling workload API", e);

      return new DefaultWorkReport(WorkStatus.FAILED);
    }

    return new DefaultWorkReport(WorkStatus.COMPLETED);
  }

  private void setTransactionOperationsPerInstance(){
      //number of operations / number of workload instances
      this.ycsbWorkload.getWorkloadConfig().setOperations(this.ycsbWorkload.getWorkloadConfig().getOperations().intValue() / this.workloadInstances.size());
  }

  private void setLoadOperationsPerInstance(int instanceCounter){

    int insertCount = this.ycsbWorkload.getWorkloadConfig().getRecordCount().intValue() / this.workloadInstances.size();
    int insertStart = instanceCounter * insertCount;


    this.ycsbWorkload.getWorkloadConfig().setInsertCount(insertCount);
    this.ycsbWorkload.getWorkloadConfig().setInsertStart(insertStart);


  }

  private void writeWorkloadSpec(int instanceCounter, String workloadSpecName){

    //Write Spec to disk
    LOGGER.debug(loggerPrefix + "Running Workload Spec: " + ycsbWorkload.toString());
    String specUri = this.evaluationContext.getEvluationRun() +   File.separator + EvaluationTask.SPECS_FOLDER
        + File.separator + instanceCounter + "_" + workloadSpecName;
    this.writeReport(specUri, GSON.toJson(ycsbWorkload));
  }



}
