package de.uulm.omi.utils;

import static com.google.common.base.Preconditions.checkNotNull;

import de.uniulm.omi.cloudiator.colosseum.client.Client;
import de.uniulm.omi.cloudiator.colosseum.client.entities.ApplicationComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.ApplicationInstance;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Instance;
import de.uniulm.omi.cloudiator.colosseum.client.entities.IpAddress;
import de.uniulm.omi.cloudiator.colosseum.client.entities.LifecycleComponent;
import de.uniulm.omi.cloudiator.colosseum.client.entities.VirtualMachine;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoint;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoints;
import java.util.List;
import java.util.Vector;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created by Daniel Seybold on 28.02.2018.
 */
public class ColosseumUtils {

  private static final Logger LOGGER = LoggerFactory.getLogger(ColosseumUtils.class);


  public static List<String> getEndpoints(Long applicationInstanceId, String networkType, DBMSNodeType dbmsNodeType){



    List<String> endpoints = new Vector<String>();

    LOGGER.debug("Looking for endpoints for ID: " + applicationInstanceId);


    Client client = ColosseumClientHelper
        .getColosseumClient();

    List<Instance> instances = client.controller(Instance.class).getList();

    List<IpAddress> ipAddresses = client.controller(IpAddress.class).getList();


    LOGGER.debug("Looking for ipAddresses of network type: " + networkType + " and DBMS node type: " + dbmsNodeType.toString());

    for (Instance instance : instances) {


      if(dbmsNodeType.equals(DBMSNodeType.ALL) || dbmsNodeType.toString().toLowerCase().equals(getComponentName(instance.getApplicationComponent()).toLowerCase())){

        //LOGGER.debug("Extracting DBMS endpoints for DBMS node type:" + dbmsNodeType.toString());
        if (instance.getApplicationInstance().equals(applicationInstanceId)) {

          for(IpAddress ipAddress : ipAddresses){
            if(ipAddress.getVirtualMachine().equals(instance.getVirtualMachine()) && ipAddress.getIpType().equals(networkType)){

              LOGGER.debug("Adding " + networkType + " ipAddress " + ipAddress.getIp() + " for applicationInstanceId " + applicationInstanceId);
              endpoints.add(ipAddress.getIp());
            }
          }

        }
      }


    }


    return endpoints;
  }

  public static String getComponentName(long applicationComponentId){

    Client client = ColosseumClientHelper
        .getColosseumClient();

    ApplicationComponent applicationComponent = client.controller(ApplicationComponent.class).get(applicationComponentId);

    checkNotNull(applicationComponent, "Did not find an applicationComponent for applicationInstanceId: " + applicationComponentId);

    LifecycleComponent component = client.controller(LifecycleComponent.class).get(applicationComponent.getComponent());

    checkNotNull(component, "Did not find a lifecycleComponent for applicationComponent: " + applicationComponent.getId());


    return  component.getName();

  }

  public static DBEndpoints extractDBEndpoints(Long appInstanceId, String networkType, DBMSNodeType dbmsNodeType){

    DBEndpoints dbEndpoints = new DBEndpoints();
    List<String> endpoints = ColosseumUtils.getEndpoints(appInstanceId, networkType, dbmsNodeType);
    if(endpoints.size() == 0){
      LOGGER.warn("Found no endpoints while fetching endpoint from Colosseum for applicationInstanceId: " + appInstanceId + " and network type: " + networkType + " node type: " + dbmsNodeType);
    }else{
      for (String endpoint: endpoints) {

        LOGGER.debug("Adding DBMS endpoint: " + endpoint);
        DBEndpoint dbEndpoint = new DBEndpoint();
        dbEndpoint.setIpAddress(endpoint);
        dbEndpoint.setPort(0);
        dbEndpoints.add(dbEndpoint);

      }

    }

    return  dbEndpoints;

  }


  public  static  boolean deleteApplicationInstance(Client colosseumClient, ApplicationInstance applicationInstance){

    // DELETE APP INSTANCE
    colosseumClient.controller(ApplicationInstance.class).delete(applicationInstance);
    // wait until instance is removed
    int tries=0; int maxtries=500;
    boolean appInstanceDeleted=false;
    while(appInstanceDeleted == false && tries <= maxtries){
      try {
        Thread.sleep(5000); //5s
        tries++;

        List<ApplicationInstance> is = colosseumClient.controller(ApplicationInstance.class).getList();
        boolean found=false;
        for(ApplicationInstance i : is){
          if(i.getId().equals(applicationInstance.getId())){
            found=true;
            break;
          }
        }
        if(!found)
          appInstanceDeleted=true;
      } catch (InterruptedException e) {
        LOGGER.error("Error while checking applicationInstance state in Colosseum!", e);
        tries = maxtries;
        return false;
      }
    }

    return appInstanceDeleted;
  }

  public static boolean deleteInstance(Client colosseumClient, Instance instance){

    // DELETE INSTANCE
    colosseumClient.controller(Instance.class).delete(instance);
    // wait until instance is removed
    int tries=0; int maxtries=500;
    boolean instanceDeleted=false;
    while(instanceDeleted == false && tries <= maxtries){
      try {
        Thread.sleep(5000); //5s
        tries++;

        List<Instance> is = colosseumClient.controller(Instance.class).getList();
        boolean found=false;
        for(Instance i : is){
          if(i.getId().equals(instance.getId())){
            found=true;
            break;
          }
        }
        if(!found)
          instanceDeleted=true;
      } catch (InterruptedException e) {
        LOGGER.error("Error while checking instance state in Colosseum!", e);
        tries = maxtries;
        return false;
      }
    }

    return instanceDeleted;
  }


  public static  boolean deleteVM(Client colosseumClient, Instance instance){

    VirtualMachine virtualMachine = colosseumClient.controller(VirtualMachine.class).get(instance.getVirtualMachine());
    colosseumClient.controller(VirtualMachine.class).delete(virtualMachine);
    // wait until virtualmachine is removed
    int tries=0;
    int maxtries=500;
    boolean vmDeleted=false;
    while(vmDeleted == false && tries <= maxtries){
      try {
        Thread.sleep(5000); //5s
        tries++;

        List<VirtualMachine> vms = colosseumClient.controller(VirtualMachine.class).getList();
        boolean found=false;
        for(VirtualMachine vm : vms){
          if(vm.getId().equals(instance.getVirtualMachine())){
            found=true;
            break;
          }
        }
        if(!found)
          vmDeleted=true;
      } catch (InterruptedException e) {
        LOGGER.error("Error while checking VM state in Colosseum!", e);
        tries = maxtries;
        return false;
      }
    }

    return vmDeleted;
  }

}
