package de.uulm.omi.utils;

import com.google.common.collect.Sets;
import de.uniulm.omi.cloudiator.colosseum.client.Client;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Cloud;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Hardware;
import de.uniulm.omi.cloudiator.colosseum.client.entities.HardwareOffer;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Image;
import de.uniulm.omi.cloudiator.colosseum.client.entities.Location;
import de.uulm.omi.cloudiator.dbms.evaluator.model.Resource.ResourceTypeEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.VM;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created by Daniel Seybold on 12.11.2018.
 */
public class TemplateHelper {

  private static final Logger LOGGER = LoggerFactory.getLogger(ColosseumUtils.class);


  public static List<VM> buildVMTemplates(Long idCloud, String locationName, String imageDistribution,  String imageVersion,
      Integer cores, Integer memory, Integer disk) throws NoSuchElementException{



    List<VM> matchingVMs = new ArrayList<>();

    //Get exactly the idCloud
    Set matchingClouds = findCloudById(idCloud);
    //Get exactly one location object to get idLocation and idParent
    Location location = findLocationByName(locationName);
    Set<String> matchingLocations = Sets.newHashSet();
    matchingLocations.add(String.valueOf(location.getId()));
    //Get the matching images
    Set matchingImages = findImage(imageDistribution, String.valueOf(imageVersion), idCloud, location.getId(), location.getParent());
    //Get the matching hardwares
    Set matchingHardware = findHardwareByRequirements(cores,memory,disk, location.getId(), location.getParent());

      //disk constraint present
      Set vmTemplatesSet = Sets.cartesianProduct(
          matchingClouds,
          matchingLocations,
          matchingImages,
          matchingHardware
      );

      LOGGER.debug("Found " + vmTemplatesSet.size() + " matching VM templates!");

      Iterator<List<String>> iterator = vmTemplatesSet.iterator();

      while (iterator.hasNext()){
        List<String> vmTemplateList = iterator.next();

        VM vm = new VM();

        vm.setResourceType(ResourceTypeEnum.VM);
        vm.setIdCloud(Integer.valueOf(vmTemplateList.get(0)));
        vm.setIdLocation(Integer.valueOf(vmTemplateList.get(1)));
        vm.setIdImage(Integer.valueOf(vmTemplateList.get(2)));
        vm.setIdHardware(Integer.valueOf(vmTemplateList.get(3)));

        matchingVMs.add(vm);

      }

      return matchingVMs;



  }

  public static Set findCloudById(Long idCloud) throws NoSuchElementException{

    Set<String> matchingClouds = Sets.newHashSet();

    Client client = ColosseumClientHelper
        .getColosseumClient();

    Cloud cloud = client.controller(Cloud.class).get(idCloud);

    if(cloud == null){
      LOGGER.warn("Did not find a matching cloud with idCloud: " + idCloud + "!");
      throw  new NoSuchElementException("Found no cloud for id: " + idCloud);
    }else {
      matchingClouds.add(String.valueOf(cloud.getId()));
    }

    return matchingClouds;

  }

  public static Set findCloudByName(String cloudName){

    Set<String> matchingClouds = Sets.newHashSet();

    Client client = ColosseumClientHelper
        .getColosseumClient();

    List<Cloud> clouds = client.controller(Cloud.class).getList();

    for (Cloud cloud : clouds) {

      if(cloud.getName().equals(cloudName))
        matchingClouds.add(String.valueOf(cloud.getId()));

    }

    if(matchingClouds.isEmpty()){
      LOGGER.warn("Did not find a matching cloud with name: " + cloudName + "!, Returning empty set!");
    }


    return matchingClouds;

  }

  public static Set findHardwareByRequirements(Integer cores, Integer memory, Integer disk,
      Long idLocation, Long idParentLocation) throws NoSuchElementException{
    Client client = ColosseumClientHelper
        .getColosseumClient();

    Set<String> matchingHardware = Sets.newHashSet();

    List<HardwareOffer> hardwareOffers = client.controller(HardwareOffer.class).getList();
    List<Hardware> hardwares = client.controller(Hardware.class).getList();

    for (HardwareOffer hardwareOffer : hardwareOffers) {
      //check if resource requirements are matching

      if(disk == null || disk.intValue()==0){
        //search only by cores and ram
        if (hardwareOffer.getNumberOfCores().intValue() == cores.intValue()
            && hardwareOffer.getMbOfRam().intValue() == memory.intValue()){

          //find associated hardware entity
          for (Hardware hardware : hardwares) {

            if(hardware.getHardwareOffer().equals(hardwareOffer.getId())
                && (hardware.getLocation().equals(idLocation) || hardware.getLocation().equals(idParentLocation))
                ){
              matchingHardware.add(String.valueOf(hardware.getId()));
            }

          }

        }
      }else{
        // search by cores, ram and disk
        if (hardwareOffer.getNumberOfCores().intValue() == cores.intValue()
            && hardwareOffer.getMbOfRam().intValue() == memory.intValue()
            && hardwareOffer.getLocalDiskSpace().intValue() == disk.intValue()){

          //find associated hardware entity
          for (Hardware hardware : hardwares) {

            if(hardware.getHardwareOffer().equals(hardwareOffer.getId())
                && (hardware.getLocation().equals(idLocation) || hardware.getLocation().equals(idParentLocation))
                ){
              matchingHardware.add(String.valueOf(hardware.getId()));
            }

          }

        }
      }

    }

    if(matchingHardware.isEmpty()){
      LOGGER.warn("Did not find a matching hardware with cores: " + cores + " and memeory: + " + memory + " and disk: + " + disk + "!");
      throw  new NoSuchElementException("Found no hardware that matches cores: " + cores + ", memory: " + memory + ", disk: " + disk
          + ",  in location: "+ idLocation + " or parentLocation: " + idParentLocation );
    }



    return matchingHardware;

  }

  public static Location findLocationByName(String locationName) throws NoSuchElementException{

    List<Location> matchingLocations = new ArrayList<>();
    Client client = ColosseumClientHelper
        .getColosseumClient();

    List<Location> locations = client.controller(Location.class).getList();

    for (Location location : locations) {

      if(location.getName().equals(locationName))
        matchingLocations.add(location);

    }


    if(matchingLocations.isEmpty()){
      LOGGER.warn("Did not find a matching location with name: " + locationName + "!");
      throw  new NoSuchElementException("Found no location that matches location name: " + locationName);
    }

    if(matchingLocations.size()>1){
      throw  new IllegalStateException("Found more than one location that matches location name: " + locationName);
    }



    return matchingLocations.get(0);

  }

  public static Set findImage(String imageName, String version, Long idCloud, Long idLocation, Long idParentLocation) throws NoSuchElementException{

    Set<String> matchingImages = Sets.newHashSet();

    Client client = ColosseumClientHelper
        .getColosseumClient();

    List<Image> images = client.controller(Image.class).getList();

    for (Image image : images) {

      if(image.getCloud().intValue() == idCloud.intValue()
          && image.getName().toLowerCase().matches(".*" + imageName.toLowerCase() +  ".*")
          && image.getName().toLowerCase().matches(".*"+ version +".*")
          && (image.getLocation().equals(idLocation) || image.getLocation().equals(idParentLocation))
          )
        matchingImages.add(String.valueOf(image.getId()));

    }

    if(matchingImages.isEmpty()){
      LOGGER.warn("Did not find a matching image with name: " + imageName + " and version: + " + version + "!, Returning empty set!");
      throw  new NoSuchElementException("Found no image that matches imageName: " + imageName + ", imageVersion: " + version
          + ", in cloud: " + idCloud + ", in location: "+ idLocation + " or parentLocation: " + idParentLocation );
    }

    return matchingImages;

  }

}
